import json
import sys
import uuid
from typing import Annotated, Optional, List, Dict, Any

import typer
from chromadb import EmbeddingFunction
from chromadb.api.models import Collection
from chromadb.utils.embedding_functions import ONNXMiniLM_L6_V2

from chroma_dp import ChromaDocument
from chroma_dp.huggingface import SupportedEmbeddingFunctions
from chroma_dp.utils.chroma import CDPUri, get_client_for_uri, remap_features


def add_to_col(col: Collection, batch: Dict[str, Any], upsert: bool = False, ef: EmbeddingFunction = None) -> None:
    try:
        if ef:
            batch["embeddings"] = ef(batch["documents"])
        if upsert:
            col.upsert(**batch)
        else:
            col.add(**batch)
    except Exception as e:
        raise e


def chroma_import(
        uri: Annotated[str, typer.Option(help="The Chroma endpoint.")],
        collection: Annotated[str, typer.Option(help="The Chroma collection.")] = None,
        inf: typer.FileText = typer.Argument(sys.stdin),
        import_file: Optional[str] = typer.Option(None, "--in", help="The Chroma collection."),
        create: Annotated[
            bool, typer.Option(help="Create the Chroma collection if it does not exist.")
        ] = False,
        limit: Annotated[int, typer.Option(help="The limit.")] = -1,
        offset: Annotated[int, typer.Option(help="The offset.")] = 0,
        batch_size: Annotated[int, typer.Option(help="The batch size.")] = 100,
        embedding_function: Optional[SupportedEmbeddingFunctions] = typer.Option(None,
                                                                                 "--ef",
                                                                                 help="The embedding function."),
        upsert: Annotated[bool, typer.Option(help="Upsert documents.")] = False,
        embed_feature: Annotated[
            str, typer.Option(help="The embedding feature.")
        ] = "embedding",
        meta_features: Annotated[
            List[str], typer.Option(help="The metadata features.")
        ] = None,
        id_feature: Annotated[str, typer.Option(help="The id feature.")] = "id",
        doc_feature: Annotated[str, typer.Option(help="The document feature.")] = "text_chunk",
):
    _embedding_function = None
    if embedding_function == SupportedEmbeddingFunctions.default:
        _embedding_function = ONNXMiniLM_L6_V2()
    if uri is None:
        raise ValueError("Please provide a ChromaDP URI.")
    parsed_uri = CDPUri.from_uri(uri)
    client = get_client_for_uri(parsed_uri)
    _collection = parsed_uri.collection or collection
    _batch_size = parsed_uri.batch_size or batch_size
    _offset = parsed_uri.offset or offset
    _limit = parsed_uri.limit or limit
    _upsert = parsed_uri.upsert or upsert
    _create = parsed_uri.create_collection or create
    _batch: Dict[str, Any] = {
        "documents": [],
        "embeddings": [],
        "metadatas": [],
        "ids": [],
    }
    chroma_collection = client.get_or_create_collection(_collection)
    lc_count = 0
    if import_file:
        with open(import_file, 'r') as inf:
            for line in inf:
                if lc_count < _offset:
                    continue
                if _limit != -1 and lc_count >= _limit:
                    break
                doc = remap_features(json.loads(line), doc_feature, embed_feature, meta_features, id_feature)
                _batch["documents"].append(doc.text_chunk)
                _batch["embeddings"].append(
                    doc.embedding if _embedding_function is None else None)  # call EF?
                _batch["metadatas"].append(doc.metadata)
                _batch["ids"].append(doc.id if doc.id else uuid.uuid4())
                if len(_batch["documents"]) >= _batch_size:
                    add_to_col(chroma_collection, _batch, _upsert, _embedding_function)
                    _batch = {
                        "documents": [],
                        "embeddings": [],
                        "metadatas": [],
                        "ids": [],
                    }
                lc_count += 1
        if len(_batch["documents"]) > 0:
            add_to_col(chroma_collection, _batch, _upsert, _embedding_function)
    else:
        for line in inf:
            if lc_count < _offset:
                continue
            if _limit != -1 and lc_count >= _limit:
                break
            doc = remap_features(json.loads(line), doc_feature, embed_feature, meta_features, id_feature)
            _batch["documents"].append(doc.text_chunk)
            _batch["embeddings"].append(
                doc.embedding if _embedding_function is None else None)  # call EF?
            _batch["metadatas"].append(doc.metadata)
            _batch["ids"].append(doc.id if doc.id else uuid.uuid4())
            if len(_batch["documents"]) >= _batch_size:
                add_to_col(chroma_collection, _batch, _upsert, _embedding_function)
                _batch = {
                    "documents": [],
                    "embeddings": [],
                    "metadatas": [],
                    "ids": [],
                }
            lc_count += 1
    if len(_batch["documents"]) > 0:
        add_to_col(chroma_collection, _batch, _upsert, _embedding_function)



import typer
from chroma_dp.chroma.chroma_export import chroma_export
from chroma_dp.chroma.chroma_import import chroma_import
from chroma_dp.transformers.embed import filter_embed
from chroma_dp.huggingface import hf_import, hf_export

app = typer.Typer(no_args_is_help=True, help="ChromaDB Data Pump commands.")

# Import commands
import_commands = typer.Typer(no_args_is_help=True, help="Import commands.")
import_commands.command(name="chroma", help="Chroma Import.", no_args_is_help=True)(chroma_import)
import_commands.command(name="hf", help="Import HuggingFace Dataset", no_args_is_help=True)(hf_import)

# Export commands
export_commands = typer.Typer(no_args_is_help=True, help="Export commands.")
export_commands.command(name="chroma", help="Imports data into ChromaDB.", no_args_is_help=True)(chroma_export)
export_commands.command(name="hf", help="Export data to HuggingFace Dataset", no_args_is_help=True)(hf_export)
# Filter commands
filter_commands = typer.Typer(no_args_is_help=True)
transform_commands = typer.Typer(no_args_is_help=True, help="Transformer commands.")
transform_commands.command(name="embed", help="Embedding Transformer. Re-embeds document contents.",
                           no_args_is_help=True)(filter_embed)

app.add_typer(export_commands, name="exp", no_args_is_help=True, help="Export Commands.")
app.add_typer(import_commands, name="imp", no_args_is_help=True, help="Import Commands.")
app.add_typer(filter_commands, name="flt", no_args_is_help=True, help="Filter commands.")
app.add_typer(transform_commands, name="tx", no_args_is_help=True, help="Filter commands.")

if __name__ == "__main__":
    app()



from typing import Optional, Dict, Any, List

import chromadb
from chromadb import ClientAPI, GetResult
from chromadb.api.models.Collection import Collection
import urllib.parse

from pydantic import BaseModel, Field

from chroma_dp import ChromaDocument


def check_collection_exists(client: ClientAPI, collection_name: str) -> bool:
    """Checks if a collection exists in ChromaDB."""
    collections = client.list_collections()
    return collection_name in [collection.name for collection in collections]


def create_collection(client: ClientAPI, collection_name: str, if_not_exist: bool = False) -> Collection:
    """Creates a collection in ChromaDB."""
    if if_not_exist:
        return client.get_or_create_collection(collection_name)
    return client.create_collection(collection_name)


def get_collection(client: ClientAPI, collection_name: str) -> Collection:
    """Gets a collection in ChromaDB."""
    return client.get_collection(collection_name)


def read_large_data_in_chunks(collection: Collection, offset: int = 0, limit: int = 100) -> GetResult:
    """Reads large data in chunks from ChromaDB."""
    result = collection.get(
        limit=limit,
        offset=offset,
        include=["embeddings", "documents", "metadatas"])
    return result


from urllib.parse import urlparse, parse_qs


class CDPUri(BaseModel):
    auth: Optional[dict] = None
    host: Optional[str] = None
    port: Optional[int] = None
    database: Optional[str] = None
    collection: Optional[str] = None
    tenant: Optional[str] = None
    batch_size: Optional[int] = None
    limit: Optional[int] = Field(None,
                                 description="Limit of documents to export. Note: "
                                             "This parameter is only valid for chroma exports")
    offset: Optional[int] = None
    create_collection: Optional[bool] = False
    upsert: Optional[bool] = False

    @staticmethod
    def from_uri(uri: str) -> "CDPUri":
        parsed = urlparse(uri)
        user_info = parsed.username or None
        password_token = parsed.password or None
        if user_info == '__auth_token__':
            auth = {
                'type': 'token',
                'token': password_token,
                'header': 'AUTHORIZATION'
            }
        elif user_info == '__x_chroma_token__':
            auth = {
                'type': 'token',
                'token': password_token,
                'header': 'X-CHROMA-TOKEN'
            }
        elif user_info is not None:
            auth = {
                'type': 'basic',
                'username': user_info,
                'password': password_token
            }
        else:
            auth = None

        host = parsed.hostname or ''
        port = parsed.port or ''

        # Splitting the path into database and collection
        path_components = parsed.path.strip('/').split('/')
        database = path_components[0] if len(path_components) > 0 else ''
        collection = path_components[1] if len(path_components) > 1 else ''

        # Parsing query parameters
        query_params = parse_qs(parsed.query)
        tenant = query_params.get('tenant', [None])[0]
        batch_size = query_params.get('batch_size', [None])[0]
        _create_collection = query_params.get('create_collection', [None])[0]
        upsert = query_params.get('upsert', [None])[0]

        return CDPUri(
            auth=auth,
            host=host,
            port=port,
            database=database,
            collection=collection,
            tenant=tenant,
            batch_size=batch_size,
            create_collection=_create_collection,
            upsert=upsert
        )


def get_client_for_uri(uri: CDPUri) -> ClientAPI:
    """Gets a ChromaDB client for a given URI."""
    client = chromadb.HttpClient(
        host=uri.host,
        port=f"{uri.port}",
        database=uri.database or chromadb.api.DEFAULT_DATABASE,
        tenant=uri.tenant or chromadb.api.DEFAULT_TENANT,
        # TODO auth
        # settings=chromadb.Settings(
        #     chroma_client_auth_provider="chromadb.auth.token.TokenAuthClientProvider",
        #     chroma_client_auth_credentials=parsed_uri.auth,
        #     chroma_client_auth_token_transport_header="X-CHROMA-TOKEN",
        # )
    )
    return client


def remap_features(in_dict: Dict[str, Any], doc_feature: str, embed_feature: str, meta_features: List[str],
                   id_feature: str) -> ChromaDocument:
    _doc = in_dict[doc_feature]
    _embed = in_dict[embed_feature] if embed_feature else None
    _meta = {k: in_dict[k] for k in meta_features} if meta_features else None
    _id = in_dict[id_feature] if id_feature else None
    return ChromaDocument(text_chunk=_doc, embedding=_embed, metadata=_meta, id=_id)



from inspect import signature
from typing import Optional, Sequence, Any, Dict, Union, Protocol, Generator, TypeVar, Iterable

from chromadb import EmbeddingFunction
from chromadb.api import ClientAPI
from chromadb.api.types import Embedding
from pydantic import BaseModel, Field, field_validator


class ChromaDocument(BaseModel):
    id: Optional[str] = Field(None, description="Document ID")
    text_chunk: Optional[str] = Field(None, description="Document text chunk")
    metadata: Optional[Dict[str, Union[str, int, float, bool]]] = Field(None, description="Document metadata")
    embedding: Optional[Embedding] = Field(None, description="Document embedding")


D = TypeVar("D", bound=ChromaDocument, contravariant=True)


class ChromaDocumentSourceGenerator(Protocol[D]):
    def __iter__(self) -> Generator[ChromaDocument, None, None]:
        ...


class ImportRequest(BaseModel):
    client: Optional[ClientAPI] = Field(None, description="ChromaDB client")
    collection: Optional[str] = Field(None, description="ChromaDB collection")
    collection_metadata: Optional[dict] = Field(None,
                                                description="Collection metadata. "
                                                            "Such as hnsw:batch_size and hnsw:sync_threshold."
                                                            "Only works if create_collection is True.")
    create_collection: Optional[bool] = Field(True, description="Create collection if it does not exist")
    document_feature: Optional[str] = Field(..., description="Document feature")
    embedding_feature: Optional[str] = Field(None, description="Embedding feature")
    metadata_features: Sequence[str] = Field([], description="Metadata features")
    id_feature: Optional[str] = Field(None, description="ID feature")
    limit: Optional[int] = Field(None, description="Limit")
    offset: Optional[int] = Field(0, description="Offset")
    batch_size: Optional[int] = Field(100, description="Batch size")
    embedding_function: Optional[Any] = Field(None, description="Embedding function")
    upsert: Optional[bool] = Field(False, description="Upsert documents, such as matched by their Id.")

    @field_validator('embedding_function')
    def check_embedding_function(cls, embedding_function):
        # Here you can add your validation logic
        # For example, check if v is an instance of your EmbeddingFunction protocol
        function_signature = signature(
            embedding_function.__class__.__call__
        ).parameters.keys()
        protocol_signature = signature(EmbeddingFunction.__call__).parameters.keys()

        if embedding_function and not function_signature == protocol_signature:
            raise ValueError(
                f"Embedding function {embedding_function} does not implement the EmbeddingFunction protocol"
            )
        return embedding_function

    class Config:
        arbitrary_types_allowed = True


class ImportResult(BaseModel):
    success: bool = Field(..., description="Whether the import was successful")


class ExportRequest(BaseModel):
    client: ClientAPI = Field(..., description="ChromaDB client")
    collection: Optional[str] = Field(..., description="ChromaDB collection")
    limit: Optional[int] = Field(-1, description="Limit")
    offset: Optional[int] = Field(0, description="Offset")
    batch_size: Optional[int] = Field(100, description="Batch size")
    embedding_feature: Optional[str] = Field(None, description="Name of the embedding feature in the dataset")
    metadata_features: Optional[Union[Sequence[str], Dict[str, str]]] = Field([],
                                                                              description="Names or mapping of metadata features")

    class Config:
        arbitrary_types_allowed = True


class ExportResult(BaseModel):
    success: bool = Field(..., description="Whether the export was successful")
    message: Optional[str] = Field(None, description="Message")



import json
import sys
from typing import Annotated, Optional, List, Dict, Any

import typer
from chromadb.utils.embedding_functions import ONNXMiniLM_L6_V2

from chroma_dp import ChromaDocument
from chroma_dp.huggingface import SupportedEmbeddingFunctions
from chroma_dp.utils.chroma import remap_features


def filter_embed(
        inf: typer.FileText = typer.Argument(sys.stdin),
        batch_size: Annotated[int, typer.Option(help="The batch size.")] = 100,
        embedding_function: Optional[SupportedEmbeddingFunctions] = typer.Option(...,
                                                                                 "--ef",
                                                                                 help="The embedding function."),
        embed_feature: Annotated[
            str, typer.Option(help="The embedding feature.")
        ] = "embedding",
        meta_features: Annotated[
            List[str], typer.Option(help="The metadata features.")
        ] = None,
        id_feature: Annotated[str, typer.Option(help="The id feature.")] = "id",
        doc_feature: Annotated[str, typer.Option(help="The document feature.")] = "text_chunk",
):
    _batch: Dict[str, Any] = {
        "documents": [],
        "embeddings": [],
        "metadatas": [],
        "ids": [],
    }
    if embedding_function == SupportedEmbeddingFunctions.default:
        _embedding_function = ONNXMiniLM_L6_V2()
    else:
        raise ValueError("Please provide a valid embedding function.")
    for line in inf:
        doc = remap_features(json.loads(line),
                             doc_feature,
                             embed_feature=embed_feature,
                             meta_features=meta_features,
                             id_feature=id_feature)
        _batch["documents"].append(doc.text_chunk)
        _batch["metadatas"].append(doc.metadata)
        _batch["ids"].append(doc.id)
        if len(_batch["documents"]) >= batch_size:
            _batch["embeddings"] = _embedding_function(_batch["documents"])
            for d, m, i, e in zip(_batch["documents"], _batch["metadatas"], _batch["ids"], _batch["embeddings"]):
                typer.echo(ChromaDocument(text_chunk=d, metadata=m, id=i, embedding=e).model_dump_json())
            _batch: Dict[str, Any] = {
                "documents": [],
                "embeddings": [],
                "metadatas": [],
                "ids": [],
            }
    if len(_batch["documents"]) > 0:
        _batch["embeddings"] = _embedding_function(_batch["documents"])
        for d, m, i, e in zip(_batch["documents"], _batch["metadatas"], _batch["ids"], _batch["embeddings"]):
            typer.echo(ChromaDocument(text_chunk=d, metadata=m, id=i, embedding=e).model_dump_json())



import datasets


def _infer_hf_type(value):
    """
    Infers the Hugging Face data type from a Python type.

    Args:
        value: The value to infer the data type from.

    Returns:
        The inferred Hugging Face data type.
    """
    if isinstance(value, bool):
        return datasets.Value("bool")
    elif isinstance(value, int):
        return datasets.Value("int32")
    elif isinstance(value, float):
        return datasets.Value("float32")
    elif isinstance(value, str):
        return datasets.Value("string")
    elif isinstance(value, list):
        if all(isinstance(elem, int) for elem in value):
            return datasets.features.Sequence(feature=datasets.Value(dtype='int32'))
        elif all(isinstance(elem, float) for elem in value):
            return datasets.features.Sequence(feature=datasets.Value(dtype='float32'))
        elif all(isinstance(elem, str) for elem in value):
            return datasets.features.Sequence(feature=datasets.Value(dtype='string'))
        else:
            raise ValueError("Unsupported list type")
    else:
        raise ValueError("Unsupported type")



import json
from typing import Annotated, Optional, List, Dict, Any
import typer
from chromadb import GetResult
from chromadb.api.models import Collection
from rich.progress import track

from chroma_dp import ChromaDocument
from chroma_dp.utils.chroma import CDPUri, get_client_for_uri


def _get_result_to_chroma_doc_list(result: GetResult) -> List[ChromaDocument]:
    """Converts a GetResult to a list of ChromaDocuments."""
    docs = []
    for idx, _ in enumerate(result["ids"]):
        docs.append(ChromaDocument(text_chunk=result["documents"][idx],
                                   embedding=result["embeddings"][idx],
                                   metadata=result["metadatas"][idx],
                                   id=result["ids"][idx]))
    return docs


def remap_features(doc: ChromaDocument,
                   doc_feature: Optional[str] = "text_chunk",
                   embed_feature: Optional[str] = "embedding",
                   id_feature: str = "id",
                   meta_features: List[str] = None) -> Dict[str, Any]:
    """Remaps ChromaDocument features to a dictionary."""

    _metas = doc.metadata if meta_features is None else {k: doc.metadata[k] for k in meta_features}
    return {
        f"{doc_feature}": doc.text_chunk,
        f"{embed_feature}": doc.embedding,
        f"{id_feature}": doc.id,
        **_metas,
    }


def read_large_data_in_chunks(collection: Collection, offset: int = 0, limit: int = 100) -> GetResult:
    """Reads large data in chunks from ChromaDB."""
    result = collection.get(
        limit=limit,
        offset=offset,
        include=["embeddings", "documents", "metadatas"])
    return result


def chroma_export(
        uri: Annotated[str, typer.Option(help="The Chroma endpoint.")],
        collection: Annotated[str, typer.Option(help="The Chroma collection.")] = None,
        export_file: Optional[str] = typer.Option(None, "--out", help="Export .jsonl file."),
        append: Annotated[bool, typer.Option(help="Append to export file.")] = False,
        limit: Annotated[int, typer.Option(help="The limit.")] = -1,
        offset: Annotated[int, typer.Option(help="The offset.")] = 0,
        batch_size: Annotated[int, typer.Option(help="The batch size.")] = 100,
        embed_feature: Annotated[
            str, typer.Option(help="The embedding feature.")
        ] = "embedding",
        meta_features: Annotated[
            List[str], typer.Option(help="The metadata features.")
        ] = None,
        id_feature: Annotated[str, typer.Option(help="The id feature.")] = "id",
        doc_feature: Annotated[str, typer.Option(help="The document feature.")] = "text_chunk",
):
    if uri is None:
        raise ValueError("Please provide a ChromaDP URI.")
    parsed_uri = CDPUri.from_uri(uri)
    client = get_client_for_uri(parsed_uri)
    _collection = parsed_uri.collection or collection
    _batch_size = parsed_uri.batch_size or batch_size
    _offset = parsed_uri.offset or offset
    _limit = parsed_uri.limit or limit
    _batch: Dict[str, Any] = {
        "documents": [],
        "embeddings": [],
        "metadatas": [],
        "ids": [],
    }
    _start = 0
    chroma_collection = client.get_collection(_collection)
    col_count = chroma_collection.count()
    total_results_to_fetch = min(col_count, _limit) if _limit > 0 else col_count
    if export_file and not append:
        with open(export_file, "w") as f:
            f.write("")
    for offset in range(_start, total_results_to_fetch, _batch_size):
        _results = _get_result_to_chroma_doc_list(read_large_data_in_chunks(chroma_collection, offset=offset, limit=min(
            total_results_to_fetch - offset, _batch_size)))
        _results = [remap_features(doc, doc_feature=doc_feature, embed_feature=embed_feature,
                                   id_feature=id_feature, meta_features=meta_features) for doc in _results]
        if export_file:
            with open(export_file, "a") as f:
                for _doc in _results:
                    f.write(json.dumps(_doc) + "\n")
        else:
            for _doc in _results:
                typer.echo(json.dumps(_doc))
